from dataclasses import dataclass
from typing import Optional

@dataclass
class ModelConfig:
    model_name: str
    local_path: str
    huggingface_id: str
    vocab_size: int
    hidden_size: int

MODEL_CONFIGS = {
    "mistral-7b": ModelConfig(
        model_name="mistral-7b",
        local_path="Mistral-7B-Instruct-v0.3",
        huggingface_id="mistralai/Mistral-7B-Instruct-v0.3",
        vocab_size=131072,
        hidden_size=4096
    ),
    "pixtral-12b": ModelConfig(
        model_name="pixtral-12b",
        local_path="pixtral-12B-2409",
        huggingface_id="mistralai/Pixtral-12B-2409",
        vocab_size=131072
    )
}

def get_model_config(model_name: str) -> Optional[ModelConfig]:
    return MODEL_CONFIGS.get(model_name)



import torch
from typing import List, Optional, Tuple

# Import the existing generate function and necessary classes
from mistral_inference.generate import sample
from mistral_sae.model import SteerableTransformer


@torch.inference_mode()
def generate_with_sae(
    encoded_prompts: List[List[int]],
    model: SteerableTransformer,
    *,
    max_tokens: int,
    temperature: float,
    eos_id: Optional[int] = None,
    sae=None,
    features=None,
    top_p: float = 0.8,
) -> Tuple[List[List[int]], List[List[float]]]:
    model.eval()
    B, _ = len(encoded_prompts), model.args.vocab_size

    # Bookkeeping
    logprobs: List[List[float]] = [[] for _ in range(B)]
    generated_tokens: List[List[int]] = [[] for _ in range(B)]
    is_finished = [False for _ in range(B)]

    max_prompt_len = max(len(p) for p in encoded_prompts)

    # Initialize cache if needed
    cache = None  # Modify if you plan to use caching

    for i in range(max_prompt_len + max_tokens):
        current_tokens = []
        current_seqlens = []

        for b in range(B):
            if i < len(encoded_prompts[b]):
                # Still processing prompt
                seq = encoded_prompts[b][: i + 1]
            elif not is_finished[b]:
                # Generating new tokens
                seq = encoded_prompts[b] + generated_tokens[b]
            else:
                # This sequence is finished
                continue

            current_tokens.extend(seq)
            current_seqlens.append(len(seq))

        if not current_tokens:
            break

        input_tensor = torch.tensor(
            current_tokens, device=model.device, dtype=torch.long
        )

        # Determine if we're processing input tokens or generating
        is_generating = i >= max_prompt_len

        # Forward pass through the model
        prelogits = model.forward(
            input_tensor,
            seqlens=current_seqlens,
            cache=cache,  # No cache in this example
            using_sae=is_generating,
            sae=sae if is_generating else None,
            features=features,
        )

        logits = torch.log_softmax(prelogits, dim=-1)

        # Process logits for each sequence
        offset = 0
        for b in range(B):
            if i < len(encoded_prompts[b]) - 1:
                # Still processing prompt, record logprob
                idx = offset + i
                next_token_id = encoded_prompts[b][i + 1]
                logprobs[b].append(logits[idx, next_token_id].item())
            elif i >= len(encoded_prompts[b]) and not is_finished[b]:
                # Generate next token
                seq_logits = logits[
                    offset + current_seqlens[b] - 1 : offset + current_seqlens[b]
                ]
                next_token = sample(seq_logits, temperature=temperature, top_p=top_p)
                generated_tokens[b].append(next_token.item())
                logprobs[b].append(seq_logits[0, next_token.item()].item())

                if eos_id is not None and next_token.item() == eos_id:
                    is_finished[b] = True

            offset += current_seqlens[b]

        if all(is_finished):
            break

    return generated_tokens, logprobs

@torch.inference_mode()
def get_input_activations_at_layer(
    encoded_prompts: List[List[int]],
    model: SteerableTransformer,
    target_layer: int,
    *,
    chunk_size: Optional[int] = None
) -> torch.Tensor:
    model = model.eval()
    with torch.no_grad():
        seqlens = [len(x) for x in encoded_prompts]

        # One chunk if size not specified
        max_prompt_len = max(seqlens)
        if chunk_size is None:
            chunk_size = max_prompt_len

        # List to store activations for the target layer
        layer_activations = []

        # Encode prompt by chunks and get activations
        for s in range(0, max_prompt_len, chunk_size):
            prompt_chunks = [p[s : s + chunk_size] for p in encoded_prompts]
            assert all(len(p) > 0 for p in prompt_chunks)

            input_ids = torch.tensor(
                sum(prompt_chunks, []), device="cuda", dtype=torch.long
            )
            chunk_seqlens = [len(p) for p in prompt_chunks]

            # Get activations for the target layer
            activations = model.get_acts(
                input_ids, chunk_seqlens, target_layer=target_layer, cache=None
            )
            layer_activations.append(activations)

        # Concatenate activations for the target layer
        all_activations = torch.cat(layer_activations, dim=0)

        return all_activations



import torch
import torch.nn as nn
from torch.nn import functional as F
import math

"""

This implementation is largely based on 'Scaling and evaluating sparse autoencoders' (Gao et al. 2024)
arXiv link: https://arxiv.org/abs/2406.04093v1

"""


class SparseAutoencoder(nn.Module):
    def __init__(
        self,
        d_model,
        d_hidden,
        k=128,
        auxk=256,
        dead_steps_threshold=10_000_000,
        data_geometric_median=None,
    ):
        super().__init__()

        self.w_enc = nn.Parameter(torch.empty(d_model, d_hidden))
        self.w_dec = nn.Parameter(torch.empty(d_hidden, d_model))

        self.b_enc = nn.Parameter(torch.zeros(d_hidden))
        self.b_pre = nn.Parameter(torch.zeros(d_model))

        self.d_model = d_model
        self.d_hidden = d_hidden
        self.k = k
        self.auxk = auxk

        batch_size = 256
        self.dead_steps_threshold = dead_steps_threshold / batch_size

        # initialize weights
        if data_geometric_median is not None:
            self.b_pre.data = torch.load(data_geometric_median)

        # tied init
        nn.init.kaiming_uniform_(self.w_enc, a=math.sqrt(5))
        self.w_dec.data = self.w_enc.data.T.clone()
        self.w_dec.data /= self.w_dec.data.norm(dim=0)

        # Initialize dead neuron tracking
        self.register_buffer(
            "stats_last_nonzero", torch.zeros(d_hidden, dtype=torch.long)
        )

    def topK_activation(self, x, k):
        topk = torch.topk(x, k=k, dim=-1, sorted=False)
        values = F.relu(topk.values)
        result = torch.zeros_like(x)
        result.scatter_(-1, topk.indices, values)
        return result

    def LN(self, x, eps=1e-5):
        mu = x.float().mean(dim=-1, keepdim=True)
        x = x - mu
        std = x.float().std(dim=-1, keepdim=True)
        x = x / (std + eps)
        return x, mu, std

    def auxk_mask_fn(self):
        dead_mask = self.stats_last_nonzero > self.dead_steps_threshold
        return dead_mask

    def forward(self, x: torch.Tensor):
        print(x.shape, self.b_pre.shape)
        x, mu, std = self.LN(x)
        x = x - self.b_pre

        pre_acts = x @ self.w_enc + self.b_enc
        latents = self.topK_activation(pre_acts, k=self.k)

        self.stats_last_nonzero *= (latents == 0).all(dim=0).long()
        self.stats_last_nonzero += 1

        dead_mask = self.auxk_mask_fn()
        num_dead = dead_mask.sum().item()

        recons = latents @ self.w_dec + self.b_pre
        recons = recons * std + mu

        if num_dead > 0:
            k_aux = min(x.shape[-1] // 2, num_dead)

            auxk_latents = torch.where(dead_mask[None], pre_acts, -torch.inf)
            auxk_acts = self.topK_activation(auxk_latents, k=k_aux)

            auxk = auxk_acts @ self.w_dec + self.b_pre
            auxk = auxk * std + mu
        else:
            auxk = None

        return recons, auxk, num_dead

    @torch.no_grad()
    def forward_val(self, x, features=None):
        x, mu, std = self.LN(x)
        x = x - self.b_pre
        pre_acts = x @ self.w_enc + self.b_enc
        latents = self.topK_activation(pre_acts, self.k)

        for feat in features:
            latents[:, feat["feat_index"]] = feat["val"]

        recons = latents @ self.w_dec + self.b_pre
        recons = recons * std + mu
        return recons

    @torch.no_grad()
    def norm_weights(self):
        self.w_dec.data /= self.w_dec.data.norm(dim=0)

    @torch.no_grad()
    def norm_grad(self):
        dot_products = torch.sum(self.w_dec.data * self.w_dec.grad, dim=0)
        self.w_dec.grad.sub_(self.w_dec.data * dot_products.unsqueeze(0))

    @torch.no_grad()
    def get_acts(self, x):
        x, mu, std = self.LN(x)
        x = x - self.b_pre
        pre_acts = x @ self.w_enc + self.b_enc
        latents = self.topK_activation(pre_acts, self.k)
        return latents



from typing import List, Optional, Union
from pathlib import Path
from dataclasses import dataclass
import torch
from mistral_inference.cache import (
    BufferCache,
    CacheInputMetadata,
)
from mistral_inference.transformer import Transformer


@dataclass
class SimpleInputMetadata:
    # rope absolute positions
    positions: torch.Tensor

    @staticmethod
    def from_seqlens(seqlens: List[int], device: torch.device) -> "SimpleInputMetadata":
        return SimpleInputMetadata(
            positions=torch.cat([torch.arange(0, seqlen) for seqlen in seqlens]).to(
                device=device, dtype=torch.long
            )
        )
    

class SteerableTransformer(Transformer):
    def __init__(self, *args, **kwargs):
        # Call the parent class's initialization
        super(SteerableTransformer, self).__init__(*args, **kwargs)

    def forward(
        self,
        input_ids: torch.Tensor,
        seqlens: List[int],
        cache: Optional[BufferCache] = None,
        using_sae: bool = False,
        sae=None,
        features=None,
    ) -> torch.Tensor:
        """
        Overrides the forward method to optionally use SAE (Sparse Autoencoder).
        If `using_sae` is True and SAE is provided, it replaces activations with SAE reconstructions.
        """
        # If using SAE, apply the custom forward logic with SAE
        if using_sae and sae is not None:
            h = self._forward_with_sae(input_ids, seqlens, cache, sae, features)
        else:
            # Otherwise, call the base class's forward method (which does not expect sae arguments)
            h = super(SteerableTransformer, self).forward_partial(
                input_ids, seqlens, cache=cache
            )

        # Handle the output, same as in the original Transformer forward
        if self.pipeline_rank < self.num_pipeline_ranks - 1:
            outs = torch.empty(
                h.shape[0], self.vocab_size, device=h.device, dtype=h.dtype
            )
        else:
            assert self.output is not None
            outs = self.output(h)

        # Broadcast output if multiple pipeline ranks are used
        if self.num_pipeline_ranks > 1:
            torch.distributed.broadcast(outs, src=self.num_pipeline_ranks - 1)

        return outs.float()

    def _forward_with_sae(
        self,
        input_ids: torch.Tensor,
        seqlens: List[int],
        cache: Optional[BufferCache],
        sae,
        features
    ) -> torch.Tensor:
        """
        Custom forward method that uses SAE (Sparse Autoencoder) for activations.
        """
        (num_toks,) = input_ids.shape
        input_metadata = SimpleInputMetadata.from_seqlens(seqlens, self.device)

        # Embedding processing for the first pipeline rank
        if self.pipeline_rank == 0:
            assert self.tok_embeddings is not None
            h = self.tok_embeddings(input_ids)
        else:
            h = torch.empty(num_toks, self.args.dim, device=self.device, dtype=self.dtype)
            torch.distributed.recv(h, src=self.pipeline_rank - 1)

        freqs_cis = self.freqs_cis[input_metadata.positions]

        # Process through the layers and apply SAE at the appropriate layer
        for local_layer_id, layer in enumerate(self.layers.values()):
            if cache is not None:
                cache_view = cache.get_view(local_layer_id, input_metadata)
            else:
                cache_view = None

            h = layer(h, freqs_cis, cache_view)

            # Apply SAE at the specified layer (assuming at all layers or specific ones)
            if sae is not None:
                with torch.no_grad():
                    h = sae.forward_val(h, features=features).to(torch.bfloat16)

        if cache is not None:
            cache.update_seqlens(seqlens)

        return h
    
    def forward_partial(
        self,
        input_ids: torch.Tensor,
        seqlens: List[int],
        cache: Optional[BufferCache] = None,
        target_layer: Optional[int] = None,
    ) -> torch.Tensor:
        assert (
            len(seqlens) <= self.args.max_batch_size
        ), f"Max batch size is {self.args.max_batch_size}, got batch size of {len(seqlens)}"
        (num_toks,) = input_ids.shape
        assert sum(seqlens) == num_toks, (sum(seqlens), num_toks)

        input_metadata: Union[CacheInputMetadata, SimpleInputMetadata]

        if cache is not None:
            input_metadata = cache.get_input_metadata(seqlens)
        else:
            input_metadata = SimpleInputMetadata.from_seqlens(seqlens, self.device)

        if self.pipeline_rank == 0:
            assert self.tok_embeddings is not None
            h = self.tok_embeddings(input_ids)
        else:
            h = torch.empty(
                num_toks, self.args.dim, device=self.device, dtype=self.dtype
            )
            torch.distributed.recv(h, src=self.pipeline_rank - 1)

        freqs_cis = self.freqs_cis[input_metadata.positions]

        for local_layer_id, layer in enumerate(self.layers.values()):
            if cache is not None:
                assert input_metadata is not None
                assert isinstance(input_metadata, CacheInputMetadata)
                cache_view = cache.get_view(local_layer_id, input_metadata)
            else:
                cache_view = None
            h = layer(h, freqs_cis, cache_view)

            global_layer_id = self.pipeline_rank * len(self.layers) + local_layer_id
            if target_layer is not None and global_layer_id == target_layer:
                return h, global_layer_id

        if cache is not None:
            cache.update_seqlens(seqlens)
        if self.pipeline_rank < self.num_pipeline_ranks - 1:
            torch.distributed.send(h, dst=self.pipeline_rank + 1)
            return h  # type: ignore
        else:
            # Last rank has a final normalization step.
            assert self.norm is not None
            return self.norm(h)  # type: ignore

    
    def get_acts(
        self,
        input_ids: torch.Tensor,
        seqlens: List[int],
        target_layer: int,
        cache: Optional[BufferCache] = None,
    ) -> torch.Tensor:
        if target_layer < 0 or target_layer >= self.n_layers:
            raise ValueError(f"Target layer must be between 0 and {self.n_layers - 1}")

        target_rank = target_layer // len(self.layers)

        if self.pipeline_rank < target_rank:
            # This rank is before the target layer, so we process normally and send the result forward
            h = self.forward_partial(input_ids, seqlens, cache)
            if self.pipeline_rank < self.num_pipeline_ranks - 1:
                torch.distributed.send(h, dst=self.pipeline_rank + 1)
            return torch.empty(
                0
            )  # Return empty tensor as this rank doesn't have the target activations

        elif self.pipeline_rank == target_rank:
            # This rank contains the target layer
            h, _ = self.forward_partial(input_ids, seqlens, cache, target_layer)
            if self.num_pipeline_ranks > 1:
                torch.distributed.broadcast(h, src=target_rank)
            return h

        else:
            # This rank is after the target layer, so we just receive the activations
            h = torch.empty(
                input_ids.shape[0], self.args.dim, device=self.device, dtype=self.dtype
            )
            torch.distributed.broadcast(h, src=target_rank)
            return h


    @staticmethod
    def from_folder(
        folder: Union[Path, str],
        max_batch_size: int = 32768,
        num_pipeline_ranks: int = 1,
        device: Union[torch.device, str] = "cuda",
        dtype: Optional[torch.dtype] = None,
    ) -> "SteerableTransformer":
        """
        Use the base class's from_folder method to load a Transformer,
        then convert the loaded model to a SteerableTransformer.
        """
        # Use super() to call the original Transformer from_folder
        model = super(SteerableTransformer, SteerableTransformer).from_folder(
            folder, max_batch_size, num_pipeline_ranks, device, dtype
        )
        
        # Convert the returned model to SteerableTransformer
        model.__class__ = SteerableTransformer

        return model


import os
import json
import zstandard as zstd
import io
from mistral_common.tokens.tokenizers.mistral import MistralTokenizer
from mistral_common.protocol.instruct.messages import UserMessage
from mistral_common.protocol.instruct.request import ChatCompletionRequest


"""

Ugly/inefficient dataloader. Would not recommend looking too hard at this.

"""


class LiveDataLoader:
    def __init__(self, batch_size, mistral_models_path):
        self.batch_size = batch_size

        folder_name = "pile-uncopyrighted/test"  # folder with jsonl.zst files
        filenames = os.listdir(folder_name)
        filenames = sorted(filenames)
        filenames = [os.path.join(folder_name, f) for f in filenames]
        self.filenames = filenames

        self.tokenizer = MistralTokenizer.from_file(
            f"{mistral_models_path}/tekken.json"
        )

        self.curr_file = 0
        self.file_path = self.filenames[self.curr_file]
        self.batch_size = batch_size
        self.file = None
        self.dctx = None
        self.stream_reader = None
        self.text_stream = None
        self._open_file()

    def _open_file(self):
        self.file = open(self.file_path, "rb")
        self.dctx = zstd.ZstdDecompressor()
        self.stream_reader = self.dctx.stream_reader(self.file)
        self.text_stream = io.TextIOWrapper(self.stream_reader, encoding="utf-8")

    def next_batch(self):
        batch = []
        for _ in range(self.batch_size * 5):
            try:
                line = next(self.text_stream)
                obj = json.loads(line.strip())
                line_text = obj["text"]
                line_text = line_text[: len(line_text) // 5]
                completion_request_u = ChatCompletionRequest(
                    messages=[UserMessage(content=line_text)]
                )
                tokens = self.tokenizer.encode_chat_completion(
                    completion_request_u
                ).tokens[:-1]
                batch.append(tokens)
            except StopIteration:
                break

        if len(batch) < self.batch_size:
            # If at end of json file, go to new one
            self.curr_file = (self.curr_file + 1) % len(self.filenames)
            self.file_path = self.filenames[self.curr_file]
            self._open_file()
            return self.next_batch()

        return batch

    def close(self):
        if self.file:
            self.file.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



import torch
from mistral_sae.liveDataloader import LiveDataLoader
from mistral_sae.generate import get_input_activations_at_layer
from mistral_sae.model import SteerableTransformer
import os

"""

Ugly/inefficient dataloader. Would not recommend looking too hard at this.

"""


class ActivationsLoader:
    def __init__(
        self,
        batch_size,
        num_batch,
        mistral_models_path,
        target_layer=24,
        dataloader_batch_size=16,
        d_model=5120,
        act_dir="pixtral-act",
    ):
        self.batch_size = batch_size
        self.num_batch = num_batch
        self.target_layer = target_layer
        self.d_model = d_model
        self.mistral_models_path = mistral_models_path

        self.tokensloader = LiveDataLoader(
            dataloader_batch_size, mistral_models_path=mistral_models_path
        )

        self.act_dir = act_dir

        if self.num_act_files() == 0:
            self.refresh_data()

        self.file_counter = 0
        filenames = os.listdir(act_dir)
        filenames = sorted(filenames)
        filenames = [os.path.join(act_dir, f) for f in filenames]
        self.filenames = filenames

    def new_data(self):
        to_return = self.get_token_acts(self.filenames[self.file_counter])
        self.file_counter += 1
        return to_return

    def num_act_files(self):
        return len(
            [
                f
                for f in os.listdir(self.act_dir)
                if os.path.isfile(os.path.join(self.act_dir, f))
            ]
        )

    def get_token_acts(self, filename):
        tokens = torch.load(filename)
        return tokens

    def refresh_data(self):
        self.delete_pt_files("pixtral-act")
        self.file_counter = 0
        self.filenames = []

        def split_list(lst, n):
            return [lst[i : i + n] for i in range(0, len(lst), n)]

        mistral7b = SteerableTransformer.from_folder(self.mistral_models_path).cuda()

        total_acts = torch.empty((0, 5120))
        for i in range(8):
            print(f"batch {i+1}")
            batch = self.tokensloader.next_batch()
            better_batch = [
                sublist for lst in batch for sublist in split_list(lst, 5120)
            ]
            for b in better_batch:
                activations = get_input_activations_at_layer(
                    [b], mistral7b, target_layer=self.target_layer
                ).cpu()
                indices_acts = torch.randperm(activations.shape[0])
                activations = activations[indices_acts]
                total_acts = torch.cat((total_acts, activations))
                del activations

        # shuffle data
        indices = torch.randperm(total_acts.shape[0])
        total_acts = total_acts[indices]
        print(total_acts.shape)

        # save data to files
        split_size = 5120
        output_dir = "pixtral-act"
        splits = torch.split(total_acts, split_size, dim=0)

        for i, split in enumerate(splits):
            file_path = os.path.join(output_dir, f"split_{i:03d}.pt")
            torch.save(split, file_path)

        filenames = os.listdir(output_dir)
        filenames = sorted(filenames)
        filenames = [os.path.join(output_dir, f) for f in filenames]
        self.filenames = filenames

        mistral7b.cpu()
        del mistral7b
        torch.cuda.empty_cache()

    def delete_pt_files(self, dir):
        for filename in os.listdir(dir):
            if filename.endswith(".pt"):
                file_path = os.path.join(dir, filename)
                try:
                    os.remove(file_path)
                except Exception as e:
                    print(f"Error deleting {file_path}: {e}")

    def needs_refresh(self):
        return self.file_counter >= self.num_act_files()
